__version__ = "1.2.0"

import logging
from collections import namedtuple

from idox.exceptions import BaseIdoxException, MalformedRequest
from idox.structs import Request
from idox.sequences import SequenceT, NumericSequence, FileSequence
from idox.idox import Idox

VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(major=1, minor=3, micro=0, releaselevel="beta", serial=0)
logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = (
    "BaseIdoxException",
    "MalformedRequest",
    "Request",
    "Idox",
    "SequenceT",
    "NumericSequence",
    "FileSequence",
)



import asyncio
import io
import sys
from enum import Enum
from pathlib import Path
from typing import Annotated

import click
from click import style as s
import typer

from idox import Idox, NumericSequence, FileSequence, SequenceT


class SequenceTypes(str, Enum):
    NUMERIC = "Numeric"


class ProtocolChoices(str, Enum):
    http = "http"
    https = "https"


class TypeChoices(str, Enum):
    get = "GET"
    post = "POST"
    head = "HEAD"
    options = "OPTIONS"
    put = "PUT"
    delete = "DELETE"
    trace = "TRACE"
    patch = "PATCH"
    connect = "CONNECT"


app = typer.Typer(pretty_exceptions_show_locals=False, no_args_is_help=True)


@app.command()
def file(
    ending_number: Annotated[
        int, typer.Argument(help="The number of requests to send up to")
    ] = None,
    request_file_path: Annotated[
        Path, typer.Option(help="The path to the burp file")
    ] = None,
    starting_number: Annotated[
        int, typer.Option(help="The base number to start at")
    ] = 0,
    numeric_step: Annotated[
        int, typer.Option(help="The step between numbers to take on each iteration")
    ] = 1,
    max_concurrency: Annotated[
        int, typer.Option(help="Maximum concurrent requests to make at a time")
    ] = 25,
    output_directory: Annotated[
        Path, typer.Option(help="Directory to store results in")
    ] = Path("./output"),
    injection_point: Annotated[
        str, typer.Option(help="The injection point to put numbers in")
    ] = "{INJECT}",
    sequence_file: Annotated[
        Path,
        typer.Option(
            help="The path to a file containing the ID's to use instead of auto incrementing id's"
        ),
    ] = None,
    protocol: ProtocolChoices = "https",
):
    sequence = get_sequencer(
        ending_number, numeric_step, sequence_file, starting_number
    )
    idox: Idox = Idox(
        sequence,
        request_file_path=request_file_path,
        protocol=protocol,
        output_directory=output_directory,
        max_concurrency=max_concurrency,
        injection_point=injection_point,
    )
    main(idox)


@app.command()
def url(
    url: Annotated[str, typer.Argument(help="The url to make requests to")],
    ending_number: Annotated[
        int, typer.Argument(help="The number of requests to send up to")
    ] = None,
    starting_number: Annotated[
        int, typer.Option(help="The base number to start at")
    ] = 0,
    numeric_step: Annotated[
        int, typer.Option(help="The step between numbers to take on each iteration")
    ] = 1,
    max_concurrency: Annotated[
        int, typer.Option(help="Maximum concurrent requests to make at a time")
    ] = 25,
    output_directory: Annotated[
        Path, typer.Option(help="Directory to store results in")
    ] = Path("./output"),
    injection_point: Annotated[
        str, typer.Option(help="The injection point to put numbers in")
    ] = "{INJECT}",
    sequence_file: Annotated[
        Path,
        typer.Option(
            help="The path to a file containing the ID's to use instead of auto incrementing id's"
        ),
    ] = None,
    request_type: TypeChoices = "GET",
    protocol: ProtocolChoices = "https",
):
    sequence = get_sequencer(
        ending_number, numeric_step, sequence_file, starting_number
    )

    idox: Idox = Idox(
        sequence,
        request_url=url,
        protocol=protocol,
        output_directory=output_directory,
        max_concurrency=max_concurrency,
        injection_point=injection_point,
        request_method=request_type.value,
    )
    main(idox)


def get_sequencer(
    ending_number: int | None,
    numeric_step: int,
    sequence_file: Path | None,
    starting_number: int,
) -> SequenceT:
    if ending_number is not None and sequence_file is not None:
        ctx = click.get_current_context()
        ctx.fail(
            "The parameters ending_number and --sequence-file cannot be used together. "
            "Please remove one and try again."
        )

    if ending_number is not None:
        sequence = NumericSequence(
            ending_number=ending_number,
            starting_number=starting_number,
            jump=numeric_step,
        )

    elif sequence_file is not None:
        if not sequence_file.exists():
            ctx = click.get_current_context()
            ctx.fail(
                "The file provided in --sequence-file does not exist. Please try again."
            )

        sequence = FileSequence(sequence_file)

    else:
        ctx = click.get_current_context()
        ctx.fail("Please provide either an ending_number or --sequence-file parameter.")

    return sequence  # noqa


def main(idox: Idox):
    async def async_main():
        click.secho("Starting Idox\r", nl=False)
        await idox.run()
        click.secho("Enumeration of target has finished\n\n", nl=False)

        output_stats = io.StringIO()
        output_stats.write(s("Statistics by response status code:\n"))
        for k, v in sorted(idox.seen_codes.items()):
            if k < 200:
                # 100's
                fg = "white"
            elif k < 300:
                # 200's
                fg = "green"
            elif k < 400:
                # 300's
                fg = "blue"
            elif k < 500:
                # 400's
                fg = "red"
            else:
                fg = "magenta"

            output_stats.write(s(k, fg=fg))
            output_stats.write(": ")
            output_stats.write(s(v, bold=True))
            output_stats.write("\n")

        output_stats.write(f"\nTotal requests made: ")
        output_stats.write(s(sum(idox.seen_codes.values()), bold=True))

        if idox.seen_errors:
            output_stats.write("\n\nSeen errors:\n")
            for k, v in sorted(idox.seen_errors.items()):
                output_stats.write(f"{k}: ")
                output_stats.write(s(v, bold=True))
                output_stats.write("\n")

            output_stats.write(f"\nTotal errors seen: ")
            output_stats.write(s(sum(idox.seen_errors.values()), bold=True))

        click.echo(output_stats.getvalue())

    asyncio.run(async_main())


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Protocol, Iterator, Any


class SequenceT(Protocol):
    iter: Iterator[Any]

    async def get_next(self) -> str:
        """Returns the next sequence to use in a request"""
        return str(next(self.iter))

    def __next__(self):
        return str(next(self.iter))

    def __iter__(self):
        return self


class NumericSequence(SequenceT):
    def __init__(self, *, ending_number: int, starting_number: int = 0, jump: int = 1):
        self.starting: int = starting_number
        self.ending: int = ending_number
        self.jump: int = jump

        self.iter = iter(range(self.starting, self.ending, self.jump))


class FileSequence(SequenceT):
    def __init__(self, file_path: Path):
        self.iter = iter(file_path.read_text().split("\n"))



import asyncio
import re
from collections import defaultdict
from enum import Enum
from pathlib import Path

import httpx
import orjson
from commons import exception_as_string

from idox.exceptions import MalformedRequest
from idox.sequences import SequenceT
from idox.structs import Request

disp_pattern = re.compile(r".*filename=\"[a-zA-Z0-9`; -_=\[\]]*\.(.*)\"")


class InjectionPoint(Enum):
    URL = 1
    BODY = 2
    COOKIES = 3
    HEADERS = 4


# noinspection PyMethodMayBeStatic
class Idox:
    def __init__(
        self,
        sequencer: SequenceT,
        *,
        max_concurrency: int = 25,
        output_directory: Path = Path("./output"),
        request_file_path: Path | None = None,
        request_url: str | None = None,
        request_method: str = "GET",
        injection_point: str = "{INJECT}",
        protocol: str = "https",
    ):
        self.output_directory: Path = output_directory
        self.max_concurrent: int = max_concurrency
        self.semaphore = asyncio.Semaphore(max_concurrency)
        self.sequencer: SequenceT = sequencer
        self.protocol: str = protocol

        self.seen_codes: dict[int, int] = defaultdict(lambda: 0)
        self.seen_errors: dict[str, int] = defaultdict(lambda: 0)

        if request_url is None and request_file_path is None:
            raise ValueError(
                "I require at-least one of incoming_request or"
                " request_file_path to work"
            )

        if request_url is None:
            with open(request_file_path, "r") as f:
                incoming_request = f.read()

            self.request: Request = self.split_request(incoming_request)

        else:
            self.request: Request = Request(
                url=request_url,
                method=request_method,
                body="",
                cookies=[],
                headers={},
            )

        # Find the injection point
        self._injection_string: str = injection_point
        if injection_point in self.request.url:
            self._injection_point: InjectionPoint = InjectionPoint.URL
        elif injection_point in str(self.request.headers):
            self._injection_point: InjectionPoint = InjectionPoint.HEADERS
        elif injection_point in str(self.request.cookies):
            self._injection_point: InjectionPoint = InjectionPoint.COOKIES
        elif injection_point in str(self.request.body):
            self._injection_point: InjectionPoint = InjectionPoint.BODY
        else:
            raise ValueError(f"Failed to find {self._injection_string} in the request")

    @classmethod
    def split_request(cls, request: str) -> Request:
        """Given a valid HTTP request, return the headers

        Returns
        -------
        Request
            The request as split

        Raises
        ------
        MalformedRequest
            The request did not match the expected HTTP spec
        """
        try:
            # Handle input request
            code = request.replace("\r\n", "\n")
            try:
                raw_headers, body = code.split("\n\n")
            except ValueError:
                # Likely missing an extra line just assume body is empty
                raw_headers, body = code.removesuffix("\n"), ""

            headers: list[str] = raw_headers.split("\n")  # type: ignore
            request_type, uri, _ = headers.pop(0).split(" ")
            request_type: str = request_type.upper()
            header_jar: dict[str, str] = {}
            for line in headers:
                k, v = line.split(": ", maxsplit=1)
                header_jar[k] = v

            cookies: list[tuple[str, str]] = []
            raw_cookies = header_jar.get("Cookie", "")
            for cookie in raw_cookies.split(";"):
                if "=" not in cookie:
                    continue
                k, v = cookie.split("=", maxsplit=1)
                cookies.append((k.strip(), v))

            if request_type == "POST":
                is_json = "json" in header_jar.get("Content-Type", "")
                if is_json:
                    body = orjson.loads(body)

            url = header_jar["Host"] + uri
            return Request(
                url=url,
                cookies=cookies,
                headers=header_jar,
                body=body,
                method=request_type,
            )

        except Exception as e:
            raise MalformedRequest from e

    def extension_from_response(self, response: httpx.Response):
        content = response.text
        content_type = response.headers.get("Content-Type", "")
        # TODO Add more extensions
        if "<!DOCTYPE html>" in content or "</html>" in content:
            extension = "html"

        elif content.startswith("<?xml"):
            extension = "xml"

        elif (
            "X5O!P%@AP[4\PZX54(P^)7CC)7}$EICAR-STANDARD-ANTIVIRUS-TEST-FILE!$H+H*"
            in content
        ):
            # Eicar me up
            extension = "txt"

        elif "image/png" in content_type:
            extension = "png"

        elif "image/jpeg" in content_type:
            extension = "jpeg"

        elif "application/pdf" in content_type:
            extension = "pdf"

        elif "application/json" in content_type:
            extension = "json"

        else:
            # Let's look at the content disposition now
            # we are passed the shit I threw at the site
            content_disp = response.headers.get("Content-Disposition")
            if content_disp is None:
                return "txt"

            try:
                extension_find = disp_pattern.search(content_disp)
                extension = extension_find.group(1).lstrip(".")
            except (ValueError, AttributeError, TypeError) as e:
                extension = "txt"
                print(exception_as_string(e))

        return extension

    async def _make_request(self, ac: httpx.AsyncClient, current_iter: str):
        url = self.request.url
        if self._injection_point is InjectionPoint.URL:
            url = self.request.url.replace(self._injection_string, current_iter)

        # Guesstimate this is correct
        if not url.startswith("http"):
            url = self.protocol + "://" + url

        try:
            async with self.semaphore:
                if self.request.body:
                    if isinstance(self.request.body, dict):
                        body = self.request.body
                        if self._injection_point is InjectionPoint.BODY:
                            body = orjson.dumps(body)
                            body = body.replace(self._injection_string, current_iter)
                            body = orjson.loads(body)

                        resp = await ac.request(
                            self.request.method,
                            url=url,
                            cookies=self.request.cookies,
                            json=body,
                            headers=self.request.headers,
                        )
                    else:
                        body = self.request.body
                        if self._injection_point is InjectionPoint.BODY:
                            body = body.replace(self._injection_string, current_iter)

                        resp = await ac.request(
                            self.request.method,
                            url=url,
                            cookies=self.request.cookies,
                            content=body,
                            headers=self.request.headers,
                        )
                else:
                    resp = await ac.request(
                        self.request.method,
                        url=url,
                        cookies=self.request.cookies,
                        headers=self.request.headers,
                    )
        except Exception as e:
            self.seen_errors[e.__class__.__name__] += 1
            return

        ext = self.extension_from_response(resp)
        ext_path = self.output_directory / ext
        all_path = self.output_directory / "all"
        code_path = self.output_directory / "status_code" / str(resp.status_code)
        ext_path.mkdir(parents=True, exist_ok=True)
        all_path.mkdir(parents=True, exist_ok=True)
        code_path.mkdir(parents=True, exist_ok=True)
        byte_content = resp.content
        output_name = f"{current_iter}.{ext}"
        base_all_path = (all_path / output_name).absolute()
        with open(base_all_path, "wb") as f:
            f.write(byte_content)

        r_1 = (ext_path / output_name).absolute()
        r_1.unlink(missing_ok=True)
        r_1.hardlink_to(base_all_path)
        r_2 = (code_path / output_name).absolute()
        r_2.unlink(missing_ok=True)
        r_2.hardlink_to(base_all_path)

        self.seen_codes[resp.status_code] += 1

    async def run(self):
        limits = httpx.Limits(
            max_keepalive_connections=None,
            max_connections=None,
            keepalive_expiry=None,
        )
        async with httpx.AsyncClient(limits=limits) as client:
            coros = [self._make_request(client, i) for i in self.sequencer]
            await asyncio.gather(*coros)



class BaseIdoxException(Exception):
    """Base exception class"""

    def __init__(self, *args):
        self.message = args[0] if args else self.__doc__

    def __str__(self):
        return self.message


class MalformedRequest(BaseIdoxException):
    """The provided request did not match the excepted HTTP format."""
